"""
level 0: site
level 1 :subject
level 2 :visit
level 3 :form
level 4 :ig
level 5 :item (variable, item_id, question)
"""

class Node:

    def __init__(self, tree, parent_index=None, level=0):
        self.children_index: set[int] = set()
        self.parent_index = parent_index
        self.values = dict()
        self.level = level
        self.tree = tree
        self.index = -1

    def add_child(self, index):
        self.children_index.add(index)

    def has_child(self) -> bool:
        return len(self.children_index) != 0

    def has_parent(self) -> bool:
        return self.parent_index is not None

    def set(self, key, value):
        self.values[key] = value

    def get(self, key):
        return self.values.get(key)

    def children(self) -> list:
        if not self.has_child():
            return []
        return [self.tree.nodes[self.level + 1][i] for i in self.children_index]

    '''
    搜寻特定的子节点
    '''

    def find_child(self, fuzzy: bool = False, **kwargs):
        if not self.has_child():
            return None
        for i in self.children_index:
            node = self.tree.nodes[self.level + 1][i]
            if fuzzy:
                if node.fuzzy_matching(**kwargs):
                    return node
            else:
                if node.satisfy(**kwargs):
                    return node
        return None

    def get_child(self, index):
        if index in self.children_index:
            return self.tree.nodes[self.level + 1][index]
        return None

    def parent(self):
        if not self.has_parent():
            return None
        return self.tree.nodes[self.level - 1][self.parent_index]

    """
    判断两个节点是否相同
    """

    def eql(self, node):
        return self.satisfy(**node.values)

    def satisfy(self, **node_values) -> bool:
        """
        根据传入的参数来判断该节点是否满足条件
        :param node_values: 需要比较的值
        :type node_values:
        :return: True 全部匹配 否则 False
        :rtype: bool
        """
        for k, v in node_values.items():
            if self.values[k] != v:
                return False
        return True

    def fuzzy_matching(self, **node_values) -> bool:
        """
        模糊匹配判断是否满足条件
        :param node_values:
        :type node_values:
        :return: True 只要满足一个，
        :rtype: bool
        """
        for k, v in node_values.items():
            if self.values[k] == v:
                return True
        return False

    @classmethod
    def new(cls, tree, parent_index=None, level=0, **node_values):
        __node = Node(tree, parent_index, level)
        for key, value in node_values.items(): __node.set(key, value)
        return __node



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class AuditRecord(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/AuditRecord+Element
    """
    
    EditPoint = Model.Attribute()
    
    UsedMethod = Model.Attribute()
    
    
    UserRef = Model.OneElement()
    
    LocationRef = Model.OneElement()
    
    DateTimeStamp = Model.OneElement()
    
    ReasonForChange = Model.OneElement()
    
    SourceID = Model.OneElement()
    



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class AbsoluteTimingConstraint(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/AbsoluteTimingConstraint
    """
    
    OID = Model.Attribute()
    
    Name = Model.Attribute()
    
    StudyEventGroupOID = Model.Attribute()
    
    StudyEventOID = Model.Attribute()
    
    TimepointTarget = Model.Attribute()
    
    TimepointPreWindow = Model.Attribute()
    
    TimepointPostWindow = Model.Attribute()
    
    
    Description = Model.OneElement()
    



import pathlib

from pyodm.core.source.path_source import PathSource
from pyodm.core.xml.reader.data.xml_data_reader import XMLDataReader
from pyodm.core.xml.reader.configuration.cdisc_configuration_specification_reader import \
    CdiscConfigurationSpecificationReader
from pyodm.factory.abstract_cdisc_xml_factory import AbstractCdiscXMLFactory
from pyodm.model import odm_specification_description


class CdiscXMLSpecificationFactory(AbstractCdiscXMLFactory):
    """
    Class 类型的对象定义在Specification 中
    """

    def default_description_files(self) -> list:
        return odm_specification_description()

    def __init__(self, data_file: PathSource, specification_files: list = None):
        """
        :param data_file: 数据文件路径
        :type data_file: pathlib.Path
        :param specification_files: 定义 Class 的配置, 默认使用内置的 ODM 配置
        :type specification_files: list[pathlib.Path]
        """
        super().__init__(data_file, specification_files)
        self._specification_files = specification_files if specification_files else self.default_description_files()

    def clazz_reader(self):
        """
        使用 Specification 的方式 解析 ODM Class 信息
        :return:
        :rtype:
        """
        CdiscConfigurationSpecificationReader().load_cdisc_definition(self, self._specification_files)






import pyodm.model.meta.cdisc_odm_entity as Meta
import pyodm.model.definition as Model


class Annotation(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/Annotation
    """
    
    SeqNum = Model.Attribute()
    
    TransactionType = Model.Attribute()
    
    ID = Model.Attribute()

    Comment = Model.OneElement()
    
    Coding = Model.ManyElements()
    
    Flag = Model.ManyElements()
    



import pathlib

from pyodm.core.support.abstract_data_source import AbstractDataSource


class PathSource(AbstractDataSource):

    def __init__(self, *paths):
        super().__init__()
        self._paths = paths

    def read(self) -> pathlib.Path:
        path = pathlib.Path()
        for value in self._paths:
            path = path.joinpath(value)
        return path



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class RangeCheck(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/RangeCheck
    """
    
    Comparator = Model.Attribute()
    
    SoftHard = Model.Attribute()
    
    ItemOID = Model.Attribute()
    
    
    ErrorMessage = Model.ManyElements()
    



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class Protocol(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/Protocol
    """
    
    
    Description = Model.OneElement()
    
    StudySummary = Model.OneElement()
    
    StudyStructure = Model.OneElement()
    
    TrialPhase = Model.OneElement()
    
    StudyTimings = Model.OneElement()
    
    StudyIndications = Model.OneElement()
    
    StudyInterventions = Model.OneElement()
    
    StudyObjectives = Model.OneElement()
    
    StudyEndPoints = Model.OneElement()
    
    StudyTargetPopulation = Model.OneElement()
    
    StudyEstimands = Model.OneElement()
    
    InclusionExclusionCriteria = Model.OneElement()
    
    StudyEventGroupRef = Model.ManyElements()
    
    WorkflowRef = Model.OneElement()
    
    Alias = Model.ManyElements()
    



from pathlib import Path

import pytest


from pyodm.core.xml.writer.cdisc_xml_writer import CdiscXmlWriter
from pyodm.exceptions import XmlWriterException
from pyodm.factory.cdisc_xml_specification_factory import CdiscXMLSpecificationFactory
from pyodm.factory.cdsic_xml_xsd_factory import CdiscXMLXsdFactory
from pyodm.model import odm_xsd_description, odm_specification_description
from pyodm.model.v2.specification import Specification



def test_case1():
    file_path = "D:\\github\\pyodm\\src\\pyodm\\unit_tests\\test_cdisc_writer\\resources\\test_case1.xml"
    cdisc = CdiscXMLXsdFactory(data_file=Path(file_path), xsd_files=odm_xsd_description())
    cw = CdiscXmlWriter(cdisc.odm(), out_put=Path("test.xml"))
    cw.write()
    assert_case1("test.xml")


def test_case2():
    file_path = "D:\\github\\pyodm\\src\\pyodm\\unit_tests\\test_cdisc_writer\\resources\\test_case2.xml"
    p = CdiscXMLXsdFactory(data_file=Path(file_path), xsd_files=odm_xsd_description())
    cw = CdiscXmlWriter(p.odm(),out_put=Path("test.xml"))
    cw.write()
    assert_case2("test.xml")


def test_case3():
    file_path = "D:\\github\\pyodm\\src\\pyodm\\unit_tests\\test_cdisc_writer\\resources\\test_case3.xml"
    p = CdiscXMLXsdFactory(data_file=Path(file_path), xsd_files=odm_xsd_description())
    cw = CdiscXmlWriter(p.odm(),out_put=Path("test.xml"))
    cw.write()
    assert_case3("test.xml")


def test_case4():
    file_path = "D:\\github\\pyodm\\src\\pyodm\\unit_tests\\test_cdisc_writer\\resources\\test_case4.xml"
    p = CdiscXMLXsdFactory(data_file=Path(file_path), xsd_files=odm_xsd_description())
    cw = CdiscXmlWriter(p.odm(),out_put=Path("test.xml"))
    cw.write()
    assert_case4("test.xml")


def test_case5():
    with pytest.raises(XmlWriterException):
        file_path = "D:\\github\\pyodm\\src\\pyodm\\unit_tests\\test_cdisc_writer\\resources\\test_case4.xml"
        cdisc = CdiscXMLSpecificationFactory(data_file=Path(file_path), specification_files=odm_specification_description())
        c = cdisc.odm().ClinicalData.first().SubjectData.first()
        cw = CdiscXmlWriter(c,out_put=Path("test.xml"))
        cw.write()


def assert_case1(output_name):
    file_path = f"D:\\github\\pyodm\\src\\pyodm\\unit_tests\\test_cdisc_writer\\{output_name}"
    p = CdiscXMLXsdFactory(data_file=Path(file_path), xsd_files=odm_xsd_description())
    x = p.odm().ClinicalData
    assert x.first().StudyOID.value == '11'
    assert x.index(1).StudyOID.value == "EX001"


def assert_case2(output_name):
    file_path = f"D:\\github\\pyodm\\src\\pyodm\\unit_tests\\test_cdisc_writer\\{output_name}"
    p = CdiscXMLXsdFactory(data_file=Path(file_path), xsd_files=odm_xsd_description())
    x = p.odm().ClinicalData.first().SubjectData.first().StudyEventData.first().ItemGroupData.first().ItemData.first().Value
    assert x.first().value == '2200-10-01'
    query = p.odm().ClinicalData.first().SubjectData.first().StudyEventData.first().ItemGroupData.first().ItemData.find(
        ItemOID="VISDT").Query.find(OID='A529A2F2-F896-4AF7-AD4D-11B7110727BC')
    assert query.Value.value == 'Value 2200-10-01 is in the future, please correct'
    assert query.AuditRecord.first().DateTimeStamp.value == '2021-03-10T13:36:51.668-00:00'


def assert_case3(output_name):
    file = f"D:\\github\\pyodm\\src\\pyodm\\unit_tests\\test_cdisc_writer\\{output_name}"
    cdisc = CdiscXMLSpecificationFactory(data_file=Path(file), specification_files=odm_specification_description())
    subject = cdisc.odm().ClinicalData.first().SubjectData.first()
    item_group_data = subject.StudyEventData.first().ItemGroupData.find(ItemGroupOID="IG.HEMATOLOGY")
    item_group_data1 = item_group_data.ItemGroupData.index(0)
    assert item_group_data1.ItemGroupOID.value == "IG.WBC"
    item_data = item_group_data1.ItemData.find(ItemOID="IT.WBC")
    assert item_data.Value.first().value == '5.2'
    item_data = item_group_data1.ItemData.index(1)
    assert item_data.Value.first().value == '10*3/uL'
    item_group_data2 = item_group_data.ItemGroupData.find(ItemGroupOID="IT.RBC")
    assert item_group_data2.ItemGroupOID.value == "IT.RBC"
    item_data = item_group_data2.ItemData.first()
    assert item_data.IsNull.value == "Yes"


def assert_case4(output_name):
    # cdisc = PyODM.v2(f"D:\\github\\pyodm\\src\\pyodm\\unit_tests\\test_cdisc_writer\\{output_name}")
    file_path = f"D:\\github\\pyodm\\src\\pyodm\\unit_tests\\test_cdisc_writer\\{output_name}"
    p = CdiscXMLXsdFactory(data_file=Path(file_path), xsd_files=odm_xsd_description())

    def clinical_data() -> Specification:
        return p.odm().ClinicalData

    subject = clinical_data().first().SubjectData.first()

    # subject = clinical_data().first().SubjectData.first
    assert subject.SubjectKey.value == '1001'
    # subject.StudyEventData.
    query = subject.StudyEventData.find(StudyEventOID="Visit1").ItemGroupData.first().ItemData.first().Query._array
    assert query[0].Value.value == "Value is in the future, please correct"
    for i in query[0].AuditRecord._array:
        print(i.ReasonForChange.value)



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class Location(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/Location+Element
    """
    
    OID = Model.Attribute()
    
    Name = Model.Attribute()
    
    Role = Model.Attribute()
    
    OrganizationOID = Model.Attribute()
    
    
    Description = Model.OneElement()
    
    MetaDataVersionRef = Model.ManyElements()
    
    Address = Model.ManyElements()
    
    Telecom = Model.ManyElements()
    
    Query = Model.ManyElements()
    



import enum


class CdiscDefineRequiredException(Exception): ...


class ParserException(Exception): ...


class XmlWriterException(Exception): ...


class CdiscFactoryException(Exception): ...


class ErrCode(enum.Enum):
    NoExist = "Not Exist"


class AttributeException(Exception):
    def __init__(self, attribute: str, clazz: str, err: ErrCode, expect=""):
        self.message = f'{attribute} {err.value} {expect} in {clazz}'

    def __str__(self): return self.message


class ElementException(Exception):
    def __init__(self, element: str, clazz: str, err: ErrCode, expect=""):
        self.message = f'{element} {err.value} {expect} in {clazz}'

    def __str__(self): return self.message



import pathlib

from cjen.nene.database_info import DataBaseInfo

from pyodm.core.xml.writer.cdisc_xml_writer import CdiscXmlWriter
from pyodm.factory.cdisc_listable_factory import CdiscListableFactory

from pyodm.tests.test_listable_factory.subject_data_resource import SubjectDataResource
from pyodm.utils.path_utils import PathUtils


def xsd_files():
    base = PathUtils.folder("pyodm", __file__).joinpath("model", "v2", "resources", "schema")
    xsds = ["ODM-admindata.xsd", "ODM-clinicaldata.xsd", "ODM-foundation.xsd",
            "ODM-protocol.xsd", "ODM-referencedata.xsd", "ODM-study.xsd"]
    return [base.joinpath(xsd) for xsd in xsds]


if __name__ == '__main__':
    database_info = DataBaseInfo.factory(dict(
        host="",  # 数据库主机 dev03
        port=3306,  # 数据库端口号
        user="",  # 数据库用户名
        pwd="",  # 数据库密码
        database=""  # 数据库库名 eclinical_edc_dev_846
    ))
    data_source = SubjectDataResource(database_info)
    factory = CdiscListableFactory(data_source, xsd_files())

    odm_data = factory.odm()
    cw = CdiscXmlWriter(odm_data, pathlib.Path(f"result.xml"))
    cw.write()



import pytest
from cjen.nene.collection_utils import list_eql

from pyodm.factory.xpath.direction import Direction
from pyodm.factory.xpath.xpath_node import XpathNode
from pyodm.factory.xpath.xpath_tree import XpathTree


@pytest.fixture
def test_data1():
    return {
        "A": ["A1", "A2", "A3"],
        "B": ["B1", "B2", "B3"],
        "A1": ["C1", "B", "C2"]
    }


def test_case1(test_data1):
    xtree = XpathTree()
    for key, children in test_data1.items():
        xtree.add(XpathNode(key), [XpathNode(child) for child in children])
    assert xtree.count() == 10
    for node in xtree.children_by_name("A"):
        assert node.name in ["A1", "A2", "A3"]
    for node in xtree.children_by_name("A1"):
        assert node.name in ["C1", "B", "C2"]
    for node in xtree.children_by_name("B"):
        assert node.name in ["B1", "B2", "B3"]


@pytest.fixture
def test_data2():
    return {
        "A": ["A1", "A2", "A3"],
        "B": ["B1", "B2", "B3"],
        "C": ["A", "B"]
    }


def test_case2(test_data2):
    xtree = XpathTree()
    for key, children in test_data2.items():
        xtree.add(XpathNode(key), [XpathNode(child) for child in children])
    assert xtree.count() == 9
    for node in xtree.children_by_name("A"):
        assert node.name in ["A1", "A2", "A3"]
    for node in xtree.children_by_name("B"):
        assert node.name in ["B1", "B2", "B3"]
    for node in xtree.children_by_name("C"):
        assert node.name in ["A", "B"]


@pytest.fixture
def test_data3():
    return {
        "A": ["A1", "A2", "A3"],
        "B": ["B", "B2", "B3"],
        "C": ["A", "B"]
    }


def test_case3(test_data3):
    xtree = XpathTree()
    for key, children in test_data3.items():
        xtree.add(XpathNode(key), [XpathNode(child) for child in children])
    assert xtree.count() == 8
    for node in xtree.children_by_name("A"):
        assert node.name in ["A1", "A2", "A3"]
    for node in xtree.children_by_name("B"):
        assert node.name in ["B", "B2", "B3"]
    for node in xtree.children_by_name("C"):
        assert node.name in ["A", "B"]
    print(xtree._all_path_index())


@pytest.fixture
def test_data4():
    return {
        "A": ["B", "A2"],
        "B": ["B1", "B2", "B3"],
        "B1": ["C1", "C2", "C3", "C4"],
    }


def test_case4(test_data4):
    xtree = XpathTree()
    for key, children in test_data4.items():
        xtree.add(XpathNode(key), [XpathNode(child) for child in children])
    assert list_eql(['A/B/B1/C1', 'A/B/B1/C2', 'A/B/B1/C3', 'A/B/B1/C4', 'A/B/B2', 'A/B/B3', 'A/A2'], xtree.xpath())
    # assert ['A/B/B1/C1', 'A/B/B1/C2', 'A/B/B1/C3', 'A/B/B1/C4', 'A/B/B2', 'A/B/B3', 'A/A2'] == xtree.xpath()


@pytest.fixture
def test_data5():
    return {
        "A": ["B", "A2"],
        "B": ["B", "B2", "B3"],
        "B1": ["C1", "C2", "C3", "C4"],
    }


def test_case5(test_data5):
    """
    父节点中的子节点包含父节点 "B": ["B", "B2", "B3"]
    """
    xtree = XpathTree()
    for key, children in test_data5.items():
        xtree.add(XpathNode(key), [XpathNode(child) for child in children])
    assert list_eql(['A/B/B2', 'A/B/B3', 'A/A2', 'B1/C1', 'B1/C2', 'B1/C3', 'B1/C4'], xtree.xpath())


@pytest.fixture
def test_data6():
    return {
        "A": ["B", "A2"],
        "B": ["B", "B2", "B3"],
        "B1": ["B", "C2"],
    }


def test_case6(test_data6):
    """
    子节点包含其它父节点 "B1": ["B", "C2"]
    """
    xtree = XpathTree()
    for key, children in test_data6.items():
        xtree.add(XpathNode(key), [XpathNode(child) for child in children])
    assert list_eql(['A/B/B2', 'A/B/B3', 'A/A2', 'B1/B/B2', 'B1/B/B3', 'B1/C2'], xtree.xpath())



@pytest.fixture
def test_data7():
    return {
        "A": ["B", "B"],
        "B": ["B", "B2", "B3"],
        "B1": ["B", "C2"],
    }


def test_case7(test_data7):
    """
    去重
    """
    xtree = XpathTree()
    for key, children in test_data7.items():
        xtree.add(XpathNode(key), [XpathNode(child) for child in children])
    assert list_eql(['A/B/B2', 'A/B/B3', 'B1/B/B2', 'B1/B/B3', 'B1/C2'], xtree.xpath())


@pytest.fixture
def test_data8():
    return {
        "A": ["B", "B"],
        "B": ["B", "B2", "B3"],
        "B1": ["B", "C2"],
    }


def test_case8(test_data8):
    """
    获取 XPATH 的方式
    """
    xtree = XpathTree()
    for key, children in test_data8.items():
        xtree.add(XpathNode(key), [XpathNode(child) for child in children])
    assert list_eql(['A/B/B2', 'A/B/B3'], xtree.xpath("A"))
    assert list_eql(['A/B/B3', 'B1/B/B3'], xtree.xpath("B3", Direction.TO))
    assert list_eql(['A/B/B2', 'A/B/B3', 'B1/B/B2', 'B1/B/B3'], xtree.xpath("B", Direction.INCLUDE))

    # assert ['A/B/B2', 'A/B/B3'] == xtree.xpath("A")
    # assert ['A/B/B3', 'B1/B/B3'] == xtree.xpath("B3", Direction.TO)
    # assert ['A/B/B2', 'A/B/B3', 'B1/B/B2', 'B1/B/B3'] == xtree.xpath("B", Direction.INCLUDE)



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class Coding(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/Coding
    """
    
    Code = Model.Attribute()
    
    System = Model.Attribute()
    
    SystemName = Model.Attribute()
    
    SystemVersion = Model.Attribute()
    
    Label = Model.Attribute()
    
    href = Model.Attribute()
    
    ref = Model.Attribute()
    
    CommentOID = Model.Attribute()
    
    



import abc
import pathlib

from pyodm.factory.cdisc_registry import CdiscRegistry


class AbstractConfigurationReader(abc.ABC):
    """
    读取配置文件的抽象方法定义，load_cdisc_definition()
    """
    @abc.abstractmethod
    def load_cdisc_definition(self, registry: CdiscRegistry, files: list[pathlib.Path]):
        """
        读取配置文件, 解析 ODM Class信息，并注册进CdiscRegistry对象
        :param registry: 存储 CdiscRegistry 的对象
        :type registry:
        :param files: 配置文件
        :type files:
        :return:
        :rtype:
        """
        ...



from pathlib import Path

from pyodm.core.data_loader.xml_path_loader import XmlPathLoader
from pyodm.core.xml.reader.mapper.xml_mapper_reader import XMLMapperReader
from pyodm.factory.cdisc_registry import CdiscRegistry


def CdiscSQL(sql_path: Path, mapper_path: Path):
    def __inner__(cls):
        return cdisc_sql_factory(cls, sql_path, mapper_path)

    return __inner__


def cdisc_sql_factory(clazz, sql_path: Path, mapper_path: Path):
    setattr(clazz, "sql_file", sql_path)
    setattr(clazz, "mapper_file", mapper_path)
    setattr(clazz, "isCdiscMapper", True)
    setattr(clazz, "__call__", __mapper)
    return clazz


def __mapper(self, registry: CdiscRegistry):
    if self.isCdiscMapper:
        elements = XMLMapperReader(registry, self).read(XmlPathLoader(self.mapper_file))
        return elements
    return []



import abc

from pyodm.core.xml.reader.configuration.cdisc_configuration_xsd_reader import CdiscConfigurationXsdReader
from pyodm.factory.abstract_cdisc_factory import AbstractCdiscFactory


class AbstractAssembleFactory(AbstractCdiscFactory, abc.ABC):
    """
    设置 解析 ODM Class 的方式
    通过 修改 reader_class, 不同的 reader_class 要对应不同 configuration_files
    CdiscConfigurationXsdReader(默认) - odm_xsd_description
    CdiscConfigurationSpecificationReader - odm_specification_description
    """
    def __init__(self, configuration_files: list = None):
        super().__init__()
        self._configuration_files = configuration_files if configuration_files else self.default_description_files()
        self._reader_class = CdiscConfigurationXsdReader

    @property
    def reader_class(self):
        """
        设置 解析 ODM Class的方式
        CdiscConfigurationXsdReader | CdiscConfigurationSpecificationReader
        :return:
        :rtype:
        """
        return self._reader_class

    @reader_class.setter
    def reader_class(self, clazz): self._reader_class = clazz

    def clazz_reader(self):
        """
        根据 reader_class 生成 解析 ODM Class 类，并注册 ODM Class
        :return:
        :rtype:
        """
        self._reader_class().load_cdisc_definition(self, self._configuration_files)


from functools import reduce
from typing import Union, List

from pyodm.utils.node import Node


class Forest:
    """
    Forest 为自定义的数据结构.

    它的目标是把传入的数组，转化为树形结构

    branch.

    degree.

    level.

    node_values.

    """

    def __init__(self):
        self._degree = 1
        self.nodes: list[list[Node]] = [[]]

    @property
    def degree(self):
        return self._degree

    def root(self) -> Node:
        return self.get_level_nodes(0)[0]

    @property
    def is_empty(self):
        return len(self.nodes[0]) == 0

    def __create_node(self, parent_index: int = None, level: int = 0, **node_values):
        """
        创建子节点，并返回子节点在对应层的索引值

        :param parent_index: 父节点的索引值

        :type parent_index: int

        :param level: 层级

        :type level: int

        :param node_values: 节点的值

        :type node_values:

        :return: 子节点的索引值

        :rtype: int
        """
        node = Node.new(self, parent_index, level, **node_values)
        index = self.__add_node(parent_index, level, node)
        if node.has_parent():
            node.parent().add_child(index)
        return index

    def add_branch(self, branch: list[dict]) -> None:
        """
        添加树的子节点

        :param branch: 从根节点开始的一个分支,该列表包含该分支每一个经过的节点

        :type branch: list[dict]

        :return: None
        :rtype: None
        """

        """验证新加的分支的第一个节点是否为根节点"""
        # if not self.is_empty:
        #     if not self.root().satisfy(**branch[0]): raise Exception(f'{branch[0]} is not root')
        """如果新加的分支的度数>当前树的度数"""
        while len(branch) > self.degree:
            self.nodes.append([])
            self._degree += 1
        """每个数组的第一位都表示根节点，所以每次进入函数时 parent 设置为 None"""
        parent = None
        for level in range(self.degree):
            """每一个分支深度不一样, 为避免越界"""
            if level <= len(branch) - 1:
                parent = self.__create_node(parent, level, **branch[level])

    def clean(self):
        self.nodes: list[list[Node]] = [[]]
        self._degree = 1

    def index(self, level, i) -> Node:
        return self.get_level_nodes(level)[i]

    def find(self, level: int, xrange: set,fuzzy: bool = False, **node_values) -> Union[None, Node]:
        """
        指定索引搜索满足props条件的值

        :param fuzzy: 是否模糊匹配， 默认为关闭 False
        :type fuzzy:
        :param level: 指定查询的层级
        :type level: int
        :param xrange: 查询的范围
        :type xrange:
        :param node_values: 满足条件的值
        :type node_values: dict
        :return:
        :rtype:
        """
        result = None
        if fuzzy: result = list(filter(lambda i: self.nodes[level][i].fuzzy_matching(**node_values), xrange))
        else: result = list(filter(lambda i: self.nodes[level][i].satisfy(**node_values), xrange))
        if not result: return None
        return self.nodes[level][result[0]]

    def search(self, level: int, fuzzy: bool = False, **node_values) -> list[Node]:
        """
        在指定层级搜索满足条件的节点

        :param fuzzy: 是否模糊搜索，默认是关闭为 False
        :type fuzzy: Boolean
        :param level: 指定的层级
        :type level: int
        :param node_values: 满足条件的值
        :type node_values: dict
        :return: 如果有满足的条件的值，则返回 list, 如果没有的话，则返回一个空 list
        :rtype: list[Node]
        """
        if fuzzy: return list(filter(lambda node: node.fuzzy_matching(**node_values), self.nodes[level]))
        return list(filter(lambda node: node.satisfy(**node_values), self.nodes[level]))

    def get_level_nodes(self, level: int) -> list[Node]:
        """
        获取 指定层级下的所有节点

        :param level: 指定的层级
        :type level:  int
        :return: 指定层级的节点列表
        :rtype: list[Node]
        """
        return self.nodes[level]

    def __add_node(self, parent_index: int, level: int, node: Node) -> int:
        """
        添加节点，并返回新加节点的索引值.

        如果是已经存在的节点，则直接返回已存在的索引值.

        如果是未存在的节点，则新建节点，并返回新的索引值.

        :param parent_index:
        :type parent_index:
        :param level:
        :type level:
        :param node:
        :type node:
        :return:
        :rtype:
        """
        level_nodes = self.nodes[level]

        if parent_index is None:
            # level = 0 的层级节点. 根节点
            # 因为level = 0 层级的节点，没有父节点，所以直接判断是否已存在相同节点，如果存在直接返回索引值
            for i in range(len(level_nodes)):
                if level_nodes[i].eql(node): return i
        else:
            # level != 0 的层级节点. 非根节点
            # 因为是有父节点的，所以判断有相同父节点的兄弟节点中是否已存在相同节点，如果存在直接返回索引值
            for i in self.nodes[level - 1][parent_index].children_index:
                if level_nodes[i].eql(node): return i
        # 节点为新节点,加入对应层级的节点列表，并返回索引
        node.index = len(level_nodes)
        self.nodes[level].append(node)
        return node.index

    def reset(self, branches: list[list[dict]]):
        """
        用新的分支 重置 Forest
        :param branches: 新传入的分支
        :type branches: list[list[dict]]
        :return:
        :rtype:
        """
        self.clean()
        self.nodes = [[]]
        for branch in branches: self.add_branch(branch)

    @classmethod
    def factory(cls, branch: list[dict] = None):
        """
        构建一个度数为 degree 的森林

        :param branch: 可选参数, 分支

        :type branch: list[dict]

        :return: 一个度数为 degree 的空的森林 或 有一个分支度数为 degree的森林

        :rtype: Forest
        """
        __tree = Forest()
        if branch is not None: __tree.add_branch(branch)
        return __tree

    @classmethod
    def transform(cls, branches: list[list[dict]]):
        """
        把 2维 数组转化为森林.

        该 2 维数组每一行为一个从根节点开始的一个分支

        :param branches: 2维数组
        :type branches: list[list[dict]]
        :return: 森林
        :rtype: Forest
        """
        __degree = reduce(lambda x, y: x if x > y else y, [len(branch) for branch in branches])
        __tree = Forest()
        for branch in branches:
            __tree.add_branch(branch)
        return __tree



import abc
from abc import ABC

from cjen.db_bigtangerine import DBBigTangerine

from pyodm.core.support.abstact_registration import AbstractRegistration
from pyodm.core.support.abstract_data_source import AbstractDataSource


class AbstractDataBaseSource(DBBigTangerine, AbstractDataSource, AbstractRegistration, ABC):

    def __init__(self, database_info):
        super().__init__(database_info)

    @abc.abstractmethod
    def read(self) -> list[list[dict]]: ...



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class ItemDef(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/ItemDef
    """
    
    OID = Model.Attribute()
    
    Name = Model.Attribute()
    
    DataType = Model.Attribute()
    
    Length = Model.Attribute()
    
    DisplayFormat = Model.Attribute()
    
    VariableSet = Model.Attribute()
    
    CommentOID = Model.Attribute()
    
    
    Description = Model.OneElement()
    
    Definition = Model.OneElement()
    
    Question = Model.OneElement()
    
    Prompt = Model.OneElement()
    
    CRFCompletionInstructions = Model.OneElement()
    
    ImplementationNotes = Model.OneElement()
    
    CDISCNotes = Model.OneElement()
    
    RangeCheck = Model.ManyElements()
    
    CodeListRef = Model.ManyElements()
    
    ValueListRef = Model.ManyElements()
    
    Coding = Model.ManyElements()
    
    Alias = Model.ManyElements()
    



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class InvestigatorRef(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/InvestigatorRef+Element
    """
    
    UserOID = Model.Attribute()
    
    



import importlib

from pyodm.core.support.abstract_configuration_reader import AbstractConfigurationReader
from pyodm.exceptions import CdiscDefineRequiredException

import lxml.etree as Etree

from pyodm.factory.xpath.xpath_node import XpathNode
from pyodm.model.definition import OneElement, ManyElements


class CdiscConfigurationSpecificationReader(AbstractConfigurationReader):
    """
    读取 Specifincation 配置文件
    """

    def load_cdisc_definition(self, registry, files: list[str]):
        for file in files:
            xml_tree = Etree.parse(file)
            root = xml_tree.getroot()
            cdisc = self._find_cdisc(root)
            if cdisc is None: raise CdiscDefineRequiredException(f"{file} 没有定义 CDISC 节点")
            for name, clazz_info in self._cdisc_loader(cdisc).items():
                module = importlib.import_module(clazz_info.get("modulePath"))
                clazz = getattr(module, clazz_info.get("clazz"))
                elements = []
                for attr_name, attr in vars(clazz).items():
                    if isinstance(attr, OneElement) or isinstance(attr, ManyElements):
                        elements.append(XpathNode(attr_name))
                registry.definition_tree.add(XpathNode(name), elements)

                registry.registry_cdisc(name, clazz)

    def _find_cdisc(self, element):
        if element.tag.lower() == "CDISC".lower():
            return element
        else:
            for sub_element in element:
                x = self._find_cdisc(sub_element)
                if x: return x

    def _cdisc_loader(self, cdisc):
        # result = {"root": cdisc.attrib.get("cdiscRoot")}
        result = {}
        for sub_element in cdisc:
            result[sub_element.tag] = {
                "clazz": sub_element.attrib.get("clazz"),
                "modulePath": sub_element.attrib.get("modulePath")
            }
        return result



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class Transition(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/Transition
    """
    
    OID = Model.Attribute()
    
    Name = Model.Attribute()
    
    SourceOID = Model.Attribute()
    
    TargetOID = Model.Attribute()
    
    StartConditionOID = Model.Attribute()
    
    EndConditionOID = Model.Attribute()
    
    



import pathlib

from pyodm.core.source.path_source import PathSource
from pyodm.core.xml.reader.configuration.cdisc_configuration_xsd_reader import CdiscConfigurationXsdReader
from pyodm.factory.abstract_cdisc_xml_factory import AbstractCdiscXMLFactory
from pyodm.model import odm_xsd_description


class CdiscXMLXsdFactory(AbstractCdiscXMLFactory):
    """
    Class 类型的定义在XSD文件中
    """

    def default_description_files(self) -> list:
        return odm_xsd_description()

    def __init__(self, data_file: PathSource, xsd_files: list = None):
        """
        :param data_file:  数据文件路径
        :type data_file: pathlib.Path
        :param xsd_files: XSD配置文件路径, 默认使用内置的 ODM 配置
        :type xsd_files: list[pathlib.Path]
        """
        super().__init__(data_file, xsd_files)
        self._xsd_files = xsd_files if xsd_files else self.default_description_files()

    def clazz_reader(self):
        """
        使用 XSD 的方式 解析 ODM Class 信息
        :return:
        :rtype:
        """
        CdiscConfigurationXsdReader().load_cdisc_definition(self, self._xsd_files)



from xml.etree.ElementTree import ElementTree

from cjen.mama.meta_data import MetaData
from lxml.etree import Element

from pyodm.core.support.abstract_data_loader import AbstractDataLoader
from pyodm.core.support.abstract_data_reader import AbstractDataReader
from pyodm.exceptions import AttributeException, ErrCode, ElementException
from pyodm.factory.cdisc_registry import CdiscRegistry
from pyodm.utils.class_utils import ClassUtils
from pyodm.utils.dict_utils import DictUtils
from pyodm.utils.etree_utils import EtreeUtils


class XMLMapperReader(AbstractDataReader):
    ODM_CLASS = "ODMClass"
    TEXT = "Text"

    def __init__(self, registry: CdiscRegistry, data: MetaData):
        super().__init__(registry)
        self._array = []
        self.data = data

    def read(self, resource: AbstractDataLoader):
        tree = self._load(resource)
        self._parse_element(tree.getroot(), level=1)
        return self._array

    def _parse_element(self, element: Element, level: int):
        if len(self._array) < level: self._array.append(dict())
        class_name = EtreeUtils.attrib(element, self.ODM_CLASS, strict=True)
        self._array[level - 1][class_name] = dict()
        attributes = DictUtils.filter(element.attrib, lambda key, value: key != self.ODM_CLASS and key != self.TEXT)
        for attribute_name, field in attributes.items():
            if not self._contain_attribute(attribute_name, class_name):
                raise AttributeException(attribute_name, class_name, ErrCode.NoExist)
            self._array[level - 1][class_name][attribute_name] = self.data.meta_data[field]
        if EtreeUtils.attrib(element, self.TEXT) is not None:
            self._array[level - 1][class_name][self.TEXT] = self.data.meta_data[self.TEXT]
        for sub_element in element.getchildren():
            sub_class_name = EtreeUtils.attrib(sub_element, self.ODM_CLASS, strict=True)
            if not self._contain_elements(sub_class_name, class_name):
                raise ElementException(sub_class_name, class_name, ErrCode.NoExist)
            self._parse_element(sub_element, level + 1)

    def _contain_attribute(self, attribute_name, class_name) -> bool:
        return attribute_name in ClassUtils.attributes_name(clazz=self.registry.get(class_name))

    def _contain_elements(self, element_name, class_name) -> bool:
        return element_name in ClassUtils.elements_name(clazz=self.registry.get(class_name))

    def _load(self, loader: AbstractDataLoader) -> ElementTree:
        return loader.load()



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class Leaf(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/Leaf
    """
    
    ID = Model.Attribute()
    
    xlink_href = Model.Attribute()
    
    
    Title = Model.ManyElements()
    



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class KeySet(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/KeySet
    """
    
    StudyOID = Model.Attribute()
    
    SubjectKey = Model.Attribute()
    
    MetaDataVersionOID = Model.Attribute()
    
    StudyEventOID = Model.Attribute()
    
    StudyEventRepeatKey = Model.Attribute()
    
    ItemGroupOID = Model.Attribute()
    
    ItemGroupRepeatKey = Model.Attribute()
    
    ItemOID = Model.Attribute()
    
    



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class StudyEventGroupRef(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/StudyEventGroupRef
    """
    
    StudyEventGroupOID = Model.Attribute()
    
    OrderNumber = Model.Attribute()
    
    Mandatory = Model.Attribute()
    
    CollectionExceptionConditionOID = Model.Attribute()
    
    
    Description = Model.OneElement()
    



import pathlib

from pyodm.core.xml.writer.cdisc_xml_writer import CdiscXmlWriter


def assert_case1(cdisc):
    cw = CdiscXmlWriter(cdisc.odm(), out_put=pathlib.Path("test.xml"))
    cw.write()


def assert_case2(cdisc):
    x = cdisc.odm().ClinicalData.first().SubjectData.first().StudyEventData.first().ItemGroupData.first().ItemData.first().Value
    assert x.first().value == '2200-10-01'
    query = cdisc.odm().ClinicalData.first().SubjectData.first().StudyEventData.first().ItemGroupData.first().ItemData.find(
        ItemOID="VISDT").Query.find(OID='A529A2F2-F896-4AF7-AD4D-11B7110727BC')
    assert query.Value.value == 'Value 2200-10-01 is in the future, please correct'
    assert query.AuditRecord.first().DateTimeStamp.value == '2021-03-10T13:36:51.668-00:00'
    cw = CdiscXmlWriter(cdisc.odm(), out_put=pathlib.Path("test.xml"))
    cw.write()


def assert_case3(cdisc):
    x = cdisc.odm().ClinicalData
    assert x.first().StudyOID.value == '11'
    assert x.index(1).StudyOID.value == "EX001"
    cw = CdiscXmlWriter(cdisc.odm(), out_put=pathlib.Path("test.xml"))
    cw.write()


def assert_case4(cdisc):
    subject = cdisc.odm().ClinicalData.first().SubjectData.first()
    assert subject.SubjectKey.value == '1001'
    query = subject.StudyEventData.find(StudyEventOID="Visit1").ItemGroupData.first().ItemData.first().Query._array
    assert query[0].Value.value == "Value is in the future, please correct"
    for i in query[0].AuditRecord._array:
        print(i.ReasonForChange.value)
    cw = CdiscXmlWriter(cdisc.odm(), out_put=pathlib.Path("test.xml"))
    cw.write()


def assert_case5(cdisc):
    cdisc.odm().ClinicalData.usage(lambda x: print(x.first().StudyOID.value))


def assert_case6(cdisc):
    subject = cdisc.odm().ClinicalData.first().SubjectData.first()
    item_group_data = subject.StudyEventData.first().ItemGroupData.find(ItemGroupOID="IG.HEMATOLOGY")
    item_group_data1 = item_group_data.ItemGroupData.index(0)
    assert item_group_data1.ItemGroupOID.value == "IG.WBC"
    item_data = item_group_data1.ItemData.find(ItemOID="IT.WBC")
    assert item_data.Value.first().value == '5.2'
    item_data = item_group_data1.ItemData.index(1)
    assert item_data.Value.first().value == '10*3/uL'
    item_group_data2 = item_group_data.ItemGroupData.find(ItemGroupOID="IT.RBC")
    assert item_group_data2.ItemGroupOID.value == "IT.RBC"
    item_data = item_group_data2.ItemData.first()
    assert item_data.IsNull.value == "Yes"
    cw = CdiscXmlWriter(cdisc.odm(), out_put=pathlib.Path("test.xml"))
    cw.write()




from pyodm.exceptions import ParserException
from pyodm.model.definition.cdisc_model import CdiscModel


class ManyElements(CdiscModel):
    """
    * (meaning optional, with zero or more occurrences)
    + (meaning required, with one or more occurrences)
    对应标识为*或+的子元素
    """
    def __init__(self):
        super().__init__()
        self._array = []

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def array(self) -> list:
        return self._array

    def __lshift__(self, other):
        self._array.append(other)

    def index(self, i):
        if i >= self.count: raise ParserException(f"{self.name} count = {self.count}, i({i}) should less than < {self.count}")
        return self._array[i]

    def first(self):
        """
        :return: 返回第一个节点
        :rtype: CdiscODMEntity
        """
        return self._array[0]

    def find(self, **attributes):
        """
        根据节点属性查找节点
        :param attributes: key 属性名， value 节点值
        :type attributes: **kwargs
        :return: 符合要求的节点 如果没有查询到则返回None
        :rtype: CdiscODMEntity 或 None
        """
        for entity in self.array:
            bingo = True
            for key, value in attributes.items():
                finder = getattr(entity, key, None)

                if finder is None or finder.no_use() or finder.value != value:
                    bingo = False
                    break
            if bingo: return entity
        return None

    @property
    def count(self):
        """
        返回节点数量
        :return:
        :rtype:
        """
        return len(self._array)


class OneElement(CdiscModel):
    """
    ? (meaning optional, with zero or one occurrence)
    """
    def __init__(self):
        super().__init__()
        self._name = "NoUse!"
        self._value = None

    @property
    def name(self): return self._name

    @name.setter
    def name(self, value): self._name = value

    @property
    def value(self): return self._value

    @value.setter
    def value(self, value): self._value = value

    def is_blank(self) -> bool: return self._value is None
